#include "edit.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <sys/poll.h>
#include <signal.h>

static int pipeErr = 0;//used as boolean flag for SIGPIPE errors with the filter command

void readFile(char *file, Line **head) {
	FILE *lines = fopen(file,"r");
	if(lines == NULL) {
		fprintf(stderr,"There is no file to read with the name %s.\n",file);
	}
	else {
		char temp[257];
		if(*head != NULL) {
			destroy(head);//remove what was there
		}
		Line *follow = *head;
		while(fgets(temp,257,lines)) {
			char *newLine;
			newLine  = strstr(temp,"\n");
			*(newLine+1) = '\0';
			if(*head == NULL) { //woot first
				*head = (Line*)malloc(sizeof(Line));
				strcpy((*head)->text,temp);
				(*head) -> next = NULL;
				follow = *head;
			}
			else {
				Line *here = (Line*)malloc(sizeof(Line));
				strcpy(here->text,temp);
				here -> next = NULL;
				follow -> next = here;
				follow = here;
			}
		}
		fclose(lines);//only close lines if it exists
	}
}

void writeFile(char *file, Line *head) { 
	FILE *write = fopen(file,"w");
	while(head != NULL) {
		fprintf(write,"%s\n",head->text);
		head = head->next;
	}
	fclose(write);
}

void insertLine(int lineNum, char *change, Line **head) {
	Line *new = (Line*)malloc(sizeof(Line));
	int i = 1;

	strcpy(new->text,change);
	new->next = NULL;
	if(*head == NULL) { //again first line
		*head = new;
		printf("The list has been initialized with the line just given.\n");
	} 
	else if(lineNum <= 0) {
		if(lineNum < 0) {
			fprintf(stderr,"Negative line numbers do not exist, line appended to beginning of file.\n");
		}
		new->next = *head;
		*head = new;
	}
	else {
		Line *temp;
		temp = *head;
		while(i < lineNum) {
			if(temp->next == NULL) {
				fprintf(stderr,"That line number does not exist, line appended to end of file.\n");
				break;
			}
			else {
				temp = temp->next;
				i++;
			}
		}
		new->next = temp->next;
		temp->next = new;
	}
}
/*
 * This method does not free the line being deleted, it returns it.
 * Freeing is the callers responisibility.
 */
Line *deleteLine(int lineNum, Line **head) {
	Line *temp = *head;
	if(lineNum <= 0) {
		fprintf(stderr,"Negative (or 0) line numbers cannot be deleted.\n");
	}
	else if(lineNum == 1) {//kill head
		*head = (*head)->next;
		return(temp);
	}
	else {
		int i = 1;
		Line *follow;
		int error = 0;
		while(i < lineNum) {
			if(temp->next == NULL) {//end and still not at requested line
				fprintf(stderr,"The requested line does not exist.\n");
				error = 1;
				break;
			}
			else {
				follow = temp;
				temp = temp->next;
			}
			i++;
		}
		if(!error) {
			follow->next = follow->next->next;
			return(temp);
		}
	}
	return(NULL);//if control gets outside of else
}
			
void pipeCommand(char *command, Line *head) {
	int error = 0;
	int pid;
	int pd1[2];
	int pd2[2];
	FILE* pipeout, *pipein;
	char buf[256];
	int status;

	/* Create a pipe. The parent will write to this pipe, and the
 	 * Child will read from this pipe */
	if (pipe(pd1) < 0) {
		fprintf(stderr, "Cannot open pipe\n");
		error = 1;
	}
	/* Create a pipe.  The child will write to this pipe, and the
 	 * parent will read from this pipe. */
	if (pipe(pd2) < 0) {
		fprintf(stderr, "Cannot open pipe\n");
		error = 1;
	}
	if(!error) {
		/* Create a child process */
		pid = fork();
		if (pid == 0) {
			/* I am the child */
			/* make stdout the write end and stdin the read end */
			if (dup2(pd1[0], fileno(stdin)) < 0) {
				fprintf(stderr, "Cannot dup2\n");
				error = 1;
			}
			if (!error && dup2(pd2[1], fileno(stdout)) < 0) {
				fprintf(stderr, "Cannot dup2\n");
				error = 1;
			}
			if(!error) {
				/* close all the ends the child does not use */
				close(pd1[0]);
				close(pd2[1]);
				close(pd1[1]);
				close(pd2[0]);
	
				/* execute the sort program */
				execlp(command,command, NULL);	
				exit(0);
			}
		}
		/* I am the parent */
	
		/* close the ends that we wont use */
		close(pd1[0]);
		close(pd2[1]);
	
		if ((pipeout = fdopen(pd1[1], "w")) == NULL) {
			fprintf(stderr, "Error creating stream for pipe\n");
			error = 1;
		} 
	
		if(!error) {
			struct sigaction action;//used later for handleing of the SIGPIPE
			while (head != NULL) {
				char *newLine = strstr(head->text,"\n");
				if(newLine == NULL) {//adds new-line and zero byte if newline not already there
					newLine = head->text;
					newLine += strlen(head->text);
					*newLine = '\n';
					*(newLine+1) = '\0';
				}
				fprintf(pipeout, "%s", head->text);
				head = head->next;
			}
			action.sa_handler = pipeErr_handle;
			//set-up for handling possible SIGPIPE
			sigemptyset(&action.sa_mask);
			action.sa_flags = 0;
			sigaction(SIGPIPE,&action,NULL);
			fflush(pipeout);
			if(!pipeErr) {
				fclose(pipeout);
				/* Create a stream to read from the pipe */
				pipein = fdopen(pd2[0], "r");

				/* Read from the pipe (data coming from "sort") */
				fgets(buf, 256, pipein);
				while (!feof(pipein)) {
					printf("%s", buf);
					fflush(stdout);
					fgets(buf, 256, pipein);
				}
				fclose(pipein);
		
				/* wait for the child to terminate */
				wait(&status);
				/* Make sure there wasn't a problem with the child */
				if (status < 0) {
					perror("Error executing command:");
				}
			}	
			else {
				pipeErr = 0;
				fprintf(stderr,"That command does not exist, or is not compatable with this program.\n");
			}
		}
	}
}




/*
 * There are no \n in the printf's in this method because other methods make sure the lines have \n's.
 */ 
void print(int start, int end, Line *head) {
	if(end < start) { //automatically switches if user got them reversed
		int temp = start;
		start = end;
		end = temp;
	}
	if(start == -1 && end == -1) { //special, print entire list
		while(head != NULL) {
			printf("%s",head->text);
			head = head -> next;
		}
	}
	else if(start < 0 || end < 0) {
		fprintf(stderr,"Negative line numbers cannot be printed.\n");
	}
	else {
		int control = 1;
		int error = 0;
		while(control < start) { //loop through till hit the start value
			if(head == NULL) {
				fprintf(stderr,"Those lines do not exist.\n");
				error = 1;
				break;
			}
			head = head -> next;
			control++;
		}
		if(!error) {
			while(control <= end) { //print till end line
				if(head == NULL) {
					fprintf(stderr,"The rest of the lines do not exist.\n");
					break;
				}
				printf("%s",head->text);
				head = head -> next;
				control++;
			}
		}
	}
}

void changeLine(int lineNum, char *change, Line *head) {
	if(lineNum <= 0) {
		fprintf(stderr,"Negative line numbers (or zero) cannot be changed.\n");
	}
	else {
		int i = 1;
		int error = 0;
		while(i < lineNum) {
			if(head->next == NULL) {
				fprintf(stderr,"That line does not exist, it cannot be changed.\n");
				error = 1;
				break;
			}
			else {
				head = head->next;
				i++;
			}
		}
		if(!error) {
			strcpy(head->text,change);
		}
	}
}

void move(int from, int to, Line **head) {
	if(to > from) { //modification due to implenentation through delete and insert 
		to--;
	}
	Line *temp = deleteLine(from,head);
	if(temp != NULL) {
		insertLine(to,temp->text,head);
		free(temp);
	}
}

void destroy(Line **head) {
	while(*head != NULL) {
		Line *temp = *head;
		*head = (*head)->next;
		free(temp);
	}
}

//only exists to handle pipe fails//
void pipeErr_handle(int signum) {
	pipeErr = 1;
}
